// MainDlg.h : interface of the CMainDlg class
//
/////////////////////////////////////////////////////////////////////////////

#pragma once

#include "./HintWin.h"

#ifdef _DEBUG
   #include "../__MyLib__/Diagnostics/Helper.h"
#endif

class CMainDlg : public CDialogImpl<CMainDlg> , public CWinDataExchange<CMainDlg>
{
public:
	enum { IDD = IDD_MAINDLG };

   class CEditImpl    : public CWindowImpl<CEditImpl    , CEdit    , CControlWinTraits> { DECLARE_EMPTY_MSG_MAP() };
   class CStaticImpl  : public CWindowImpl<CStaticImpl  , CStatic  , CControlWinTraits> { DECLARE_EMPTY_MSG_MAP() };
   class CComboBoxImpl: public CWindowImpl<CComboBoxImpl, CComboBox, CControlWinTraits> { DECLARE_EMPTY_MSG_MAP() };
   class CCheckBoxImpl: public CWindowImpl<CCheckBoxImpl, CButton  , CControlWinTraits> { DECLARE_EMPTY_MSG_MAP() };

private:
   CComboBoxImpl m_cbxErrCode;
   CEditImpl m_edtName;
   CEditImpl m_edtDscr;
   CStaticImpl m_statIcoWarn;
   CStaticImpl m_statIcoFinder;
   CCheckBoxImpl m_chkbxIsHex;

   CIcon m_Icon, m_IconSmall, m_IcoWin;
   CCursor m_CursorSearch;
   CPen m_RectanglePen;
   CHintWin m_HintWin;

   BOOL m_bStartSearchWindow;
   HWND m_hWndFound;

   DWORD m_dwLastClickOnEditBoxDescr, m_dwLastLastClickOnEditBoxDescr;
   BOOL m_bErrCodeDescrSysMsg;
   DWORD m_dwLanguageId;

private:
	BEGIN_MSG_MAP_EX(CMainDlg)
		MSG_WM_INITDIALOG(OnInitDialog)
      MSG_WM_DESTROY(OnDestroy)
      MSG_WM_MOUSEMOVE(OnMouseMove)
      MSG_WM_LBUTTONUP(OnLButtonUp)
      MSG_WM_ACTIVATEAPP(OnActivateApp)
      MSG_WM_SETCURSOR(OnSetCursor)

      COMMAND_ID_HANDLER_EX(IDOK, OnClose)
		COMMAND_ID_HANDLER_EX(IDCANCEL, OnClose)
    //COMMAND_ID_HANDLER_EX(IDC_CBX_ERRCODE, OnErrCode)
    //COMMAND_ID_HANDLER_EX(IDC_EDIT_ERRCODE_DESCR, OnErrCodeDescr)
      COMMAND_HANDLER_EX(IDC_CBX_ERRCODE, CBN_SELCHANGE , OnSelectChangeErrCode)
      COMMAND_HANDLER_EX(IDC_CBX_ERRCODE, CBN_EDITCHANGE, OnEditChangeErrCode)
      COMMAND_HANDLER_EX(IDC_STATIC_ICON_FINDER_TOOL, STN_CLICKED, OnFinderTool)
    //COMMAND_HANDLER_EX(IDC_STATIC_ICON_FINDER_TOOL, STN_DBLCLK , OnFinderTool)
      COMMAND_HANDLER_EX(IDC_CHECK_ISHEX, BN_CLICKED , OnChangeTypeValue)
	END_MSG_MAP()
	BEGIN_DDX_MAP(CMainDlg)
		//DDX_CONTROL(IDOK                     , m_BtnOk)
      DDX_CONTROL(IDC_CBX_ERRCODE            , m_cbxErrCode)
      DDX_CONTROL(IDC_EDIT_ERRCODE_NAME      , m_edtName)
      DDX_CONTROL(IDC_EDIT_ERRCODE_DESCR     , m_edtDscr)
      DDX_CONTROL(IDC_STATIC_ICO             , m_statIcoWarn)
      DDX_CONTROL(IDC_STATIC_ICON_FINDER_TOOL, m_statIcoFinder)
      DDX_CONTROL(IDC_CHECK_ISHEX            , m_chkbxIsHex)
	END_DDX_MAP()

	BOOL OnInitDialog(HWND, LPARAM);
   void OnDestroy();
   void OnMouseMove(UINT nFlags, CPoint point) { if (m_bStartSearchWindow) OnRunFinder(); }
   void OnLButtonUp(UINT nFlags, CPoint point) { if (m_bStartSearchWindow) OnStopFinder(true); }
	void OnClose(UINT, int, CWindow);
   BOOL OnSetCursor(CWindow wnd, UINT nHitTest, UINT message);
 
 //void OnErrCode(UINT wNotifyCode, int, CWindow) { SetMsgHandled(FALSE); ::OutputDebugString(::Format(_T("   CMainDlg::OnErrCode: wNotifyCode=%i\r\n"), wNotifyCode)); }
 //void OnErrCodeDescr(UINT wNotifyCode, int, CWindow) { SetMsgHandled(FALSE); ::OutputDebugString(::Format(_T("   CMainDlg::OnErrCodeDescr: wNotifyCode=%i\r\n"), wNotifyCode)); }
   void OnEditChangeErrCode(UINT, int, CWindow);
   void OnSelectChangeErrCode(UINT, int, CWindow) { PostMessage(WM_COMMAND, MAKEWPARAM(IDC_CBX_ERRCODE, CBN_EDITCHANGE), (LPARAM)m_hWnd); } // async call OnEditChangeErrCode
   void OnFinderTool(UINT uNotifyCode, int nID, CWindow wndCtl) { OnStartFinder(); }
   void OnActivateApp(BOOL bActive, DWORD dwThreadID) { if (m_bStartSearchWindow && !bActive) OnStopFinder(false); }
   void OnChangeTypeValue(UINT uNotifyCode, int nID, CWindow wndCtl);
   void OnTripleClickDescr();
   void OnStaticDescrBttnDownClick();

   void OnStartFinder();
   void OnRunFinder(); 
   void OnStopFinder(bool bFromUser);

   BOOL CheckWindowValidity(HWND hWndToCheck);
   void RefreshFoundWindow();
   void HighlightFoundWindow();
   void DisplayInfoOnFoundWindow();

   BOOL GetErrCodeValAsDec(OUT LONG &val);
   BOOL GetErrCodeValAsHex(OUT LONG &val);

   static BOOL CALLBACK EnumChild(HWND hWndChild, LPARAM lParam) { return ((CMainDlg*)lParam)->EnumChild(hWndChild); }
   BOOL EnumChild(HWND hWndChild);
};


//#ifdef _DEBUG
//#include "./Diagnostics/Helper.h"
//
//class CMainDlgDebug : public CMainDlg {
//public:
//   virtual BOOL ProcessWindowMessage(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT& lResult, DWORD dwMsgMapID) {
//      ::OutputDebugStr(::Format(_T("CMainDlg: uMsg=[%s]\r\n"), (LPCTSTR)nsHelper::WindowMessage(uMsg)));
//      return __super::ProcessWindowMessage(hWnd, uMsg, wParam, lParam, lResult, dwMsgMapID);
//   }
//};
//#endif