////////////////////////////////////////////////////////////////////////////////
// File name: Bitmap.cpp
// Author: Sergey Krivulya (Ce p) - KSerg (SeregaLBN)
// e-mail: Sergey_Krivulya@UkrPost.Net
// ICQ: 72099167
// WWW page: http://kserg77.narod.ru/Bitmap.html
// Date: 2010 11 01
//
// Description:   BMP
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2007-2010 Sergey Krivulya
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
////////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include <Math.h>
#include "Bitmap.h"
#pragma comment(lib, "msimg32")

/** /
void logBitColor(HGDIOBJ hObj) {
   DWORD dwType = ::GetObjectType(hObj);
   HGDIOBJ hBmp = NULL;
   switch (dwType) {
      case OBJ_DC:
      case OBJ_MEMDC:
         hBmp = ::GetCurrentObject((HDC)hObj, OBJ_BITMAP);
         break;
      case OBJ_BITMAP:
         hBmp = hObj;
         break;
   }
   if (!hBmp) {
      LOG_PUT(g_pLog, Put, LL_ERROR) _T("Unknown GDI object - type %i "), dwType);
   } else {
      BITMAP bmp;
      BOOL bRes = ::GetObject(hBmp, sizeof(bmp), &bmp);
      if (bRes) {
         LOG_PUT(g_pLog, Put, LL_DEBUG) _T("GDI object has bits %i "), bmp.bmBitsPixel);
      } else {
         DWORD dwErrCode = ::GetLastError();
         LOG_PUT(g_pLog, PutErrorCode, LL_ERROR) _T("GDI object error: "), dwErrCode);
      }
   }
}
/**/

#ifndef WINDOWS_BITMAP_SIGNATURE
   #define WINDOWS_BITMAP_SIGNATURE 0x4D42 //    BMP: 0x42 = "B" 0x4d = "M"
#endif

namespace nsBmpInternal {

   namespace {
      __forceinline WORD DIBNumColors(const BITMAPINFOHEADER &bih) { // Calculates the number of entries in the color table.
         if (bih.biClrUsed) {
            return (WORD)bih.biClrUsed;
         } else {
            switch (bih.biBitCount) {
            case 1:
            case 4:
            case 8:
               return 1 << bih.biBitCount;
            default: return 0;
            }
         }
      }
      __forceinline WORD PaletteSize(const BITMAPINFOHEADER &bih) { // Calculates the number of bytes in the color table.
         return DIBNumColors(bih) * sizeof(RGBQUAD);
      }
      __forceinline BYTE* FindDIBBits(const BITMAPINFOHEADER &bih) { // Locate the image bits in a CF_DIB format DIB.
         return ((BYTE*)&bih) + bih.biSize + PaletteSize(bih);
      }
      #define WIDTHBYTES(bits)      ((((bits) + 31)>>5)<<2) // How wide, in bytes, would this many bits be, DWORD aligned?
      __forceinline DWORD BytesPerLine(const BITMAPINFOHEADER &bih) { // Calculates the number of bytes in one scan line.
         return WIDTHBYTES(bih.biWidth * bih.biPlanes * bih.biBitCount);
      }
      __forceinline DWORD DibSectionSize(const BITMAPINFOHEADER &bih) {
         if ((bih.biCompression == BI_RGB) && (bih.biSizeImage==0)) // biSizeImage may be set to zero for BI_RGB bitmaps.
            return BytesPerLine(bih) * bih.biHeight; //  
         else
            return bih.biSizeImage;                  //  ,   
      }
      __forceinline DWORD BitmapSizeInBytes(const BITMAPINFOHEADER &bih) {
         return sizeof(BITMAPINFOHEADER) + PaletteSize(bih) + DibSectionSize(bih);
      }
   }

   /** /
   BOOL FindDevMode(IN WORD dwBit, OUT DEVMODE &devMode) {
      BOOL bReturn = FALSE;
      DWORD dwErrCode = NO_ERROR;
      DISPLAY_DEVICE displayDevice = {sizeof(DISPLAY_DEVICE), };
      BOOL bRes = ::EnumDisplayDevices(
         NULL,             // LPCTSTR lpDevice,  // device name
         0,                // DWORD iDevNum,     // display device
         &displayDevice,   // device information
         0                 // DWORD dwFlags      // reserved
      );
      if (!bRes) {
         dwErrCode = ::GetLastError();
      } else {
         //LOG_PUT(g_pLog, Put, LL_DEBUG) _T(" ========================================= "));
         //LOG_PUT(g_pLog, Put, LL_DEBUG) _T(" Display device name [%s] "), displayDevice.DeviceName);
         for (int j=0; bRes; j++) {
            //DEVMODE devMode = {sizeof(DEVMODE), };
            bRes = ::EnumDisplaySettings(
               displayDevice.DeviceName,  // LPCTSTR lpszDeviceName,  // display device
               j,                         // DWORD iModeNum,          // graphics mode
               &devMode                   // graphics mode settings
            );
            if (!bRes) {
               dwErrCode = ::GetLastError();
            } else {
               // LOG_PUT(g_pLog, Put, LL_DEBUG) _T(" settings: %ix%i  bits %i "), devMode.dmPelsWidth, devMode.dmPelsHeight, devMode.dmBitsPerPel);
               if (dwBit == devMode.dmBitsPerPel) {
                  bReturn = TRUE;
                  break; // found and create
               }
            }
         }
      }
      ::SetLastError(dwErrCode);
      return bReturn;
   }

   /** /
   HDC CreateDisplayDC(WORD dwBit) {
      HDC hDC = NULL;
      DWORD dwErrCode = NO_ERROR;
      DEVMODE devMode;
      BOOL bRes = FindDevMode(dwBit, devMode);
      if (!bRes) {
         dwErrCode = ::GetLastError();
      } else {
         // LOG_PUT(g_pLog, Put, LL_DEBUG) _T(" settings: %ix%i  bits %i "), devMode.dmPelsWidth, devMode.dmPelsHeight, devMode.dmBitsPerPel);
         hDC = ::CreateDC(
            _T("DISPLAY"), // LPCTSTR lpszDriver,        // driver name
            0,             // LPCTSTR lpszDevice,        // device name
            NULL,          // LPCTSTR lpszOutput,        // not used; should be NULL
            &devMode       // optional printer data
         );
         if (!hDC) {
            DWORD dwErrCode = ::GetLastError();
            //LOG_PUT(g_pLog, PutErrorCode, LL_DEBUG) _T(" CreateDC: "), dwErrCode);
         } else {
            // LOG_PUT(g_pLog, Put, LL_DEBUG) _T(" testDC = 0x%08X "), testDC);
         }
         //::DeleteDC(hDC);
      }
      ::SetLastError(dwErrCode);
      return hDC;
   }

   HDC CreateCompatibleDC(WORD dwBit) {
      DWORD dwErrCode = NO_ERROR;
      HDC hDC = CreateDisplayDC(dwBit);
      BOOL bRes = !!hDC;
      if (!bRes) {
         dwErrCode = ::GetLastError();
         hDC = ::CreateCompatibleDC(NULL); // @TODO: ,      ...
         //if (!hDC)
         //   dwErrCode = ::GetLastError();
      } else {
         HDC hDCTmp = ::CreateCompatibleDC(hDC);
         ::DeleteDC(hDC);
         hDC = hDCTmp;
      }
      ::SetLastError(dwErrCode);
      return hDC;
   }
   /**/

   HBITMAP CreateBitmap(WORD wBitCount, int cx, int cy) {
      BITMAPINFO bmi = { // create DIB Section
         {                             // BITMAPINFOHEADER    bmiHeader;
            sizeof BITMAPINFOHEADER,   // DWORD      biSize;
            cx,                        // LONG       biWidth;
            cy,                        // LONG       biHeight;
            1,                         // WORD       biPlanes;
            wBitCount,                 // WORD       biBitCount;
            BI_RGB,                    // DWORD      biCompression;
            0,                         // DWORD      biSizeImage;
            0,                         // LONG       biXPelsPerMeter;
            0,                         // LONG       biYPelsPerMeter;
            0,                         // DWORD      biClrUsed;
            0                          // DWORD      biClrImportant;
         },
         {                             // RGBQUAD             bmiColors[1];
            {
               0,                      // BYTE    rgbBlue;
               0,                      // BYTE    rgbGreen;
               0,                      // BYTE    rgbRed;
               0                       // BYTE    rgbReserved;
            }
         }
      };

      // create valid handle bitmap
      HDC hTempDC = ::GetDC(NULL);
      void *pvData = NULL;
      HBITMAP hBmp = ::CreateDIBSection(hTempDC, &bmi, DIB_RGB_COLORS, (void**)&pvData, NULL, 0);
      DWORD dwErrCode = NO_ERROR;
      if (!hBmp) {
         dwErrCode = ::GetLastError();
      }
      ::ReleaseDC(NULL, hTempDC);
      ::SetLastError(dwErrCode);
      return hBmp;
   }

   const BITMAPINFO *toBmpInfo(IN const BYTE *pbData, IN DWORD nbSize) {
      const BITMAPINFO *pBmpI = NULL;
      DWORD dwErrCode = NO_ERROR;
      if ((pbData == NULL) || (nbSize == 0)) {
         dwErrCode = ERROR_INVALID_PARAMETER;
      } else {
         if (nbSize < (sizeof(BITMAPINFO))) {
            dwErrCode = ERROR_INVALID_DATA;
         } else {
            const BITMAPINFOHEADER *pBmpIH = (const BITMAPINFOHEADER*)pbData;
            if (pBmpIH->biSize != sizeof(BITMAPINFOHEADER)) {
               dwErrCode = ERROR_INVALID_DATA;
            } else {
               //    
               DWORD nbSizeCheck = BitmapSizeInBytes(*pBmpIH);
               if ((nbSizeCheck != nbSize) &&
                   (nbSizeCheck != nbSize-2)) //       - 2 
               {
                  _ASSERT(FALSE);
                  dwErrCode = ERROR_INVALID_DATA;
               } else {
                  pBmpI = (const BITMAPINFO*)pBmpIH;
               }
            }
         }
      }
      ::SetLastError(dwErrCode);
      return pBmpI;
   }
   SIZE SizeBitmap(IN const BYTE *pbData, IN DWORD nbSize) {
      ::SetLastError(NO_ERROR);
      SIZE size = {0,0};
      const BITMAPINFO *pBmpI = toBmpInfo(pbData, nbSize);
      if (pBmpI) {
         size.cx = pBmpI->bmiHeader.biWidth;
         size.cy = pBmpI->bmiHeader.biHeight;
      }
      return size;
   }

   HBITMAP LoadBitmap(const BITMAPINFO &bmpI) {
      HBITMAP hBmp = NULL;
      DWORD dwErrCode = NO_ERROR;
      // 1. Create bitmap
      {
         HDC hDC = ::GetDC(NULL);
         switch (bmpI.bmiHeader.biBitCount) {
         case 1:
            {
            //BITMAPINFOHEADER bih = bmpI.bmiHeader;
            //bih.biSizeImage = 0;
            //bih.biXPelsPerMeter = bih.biYPelsPerMeter = 0;
            //bih.biClrUsed = 2;
            DWORD fdwInit = 2; //    2,    API - LoadImage  LoadBitmap
            hBmp = ::CreateDIBitmap(hDC, &bmpI.bmiHeader/*&bih*/, fdwInit, NULL, &bmpI, 0);
            }
            break;
         case 4 : // hBmp = CreateBitmap( 4, bmpI.bmiHeader.biWidth, bmpI.bmiHeader.biHeight); break;
         case 8 : // hBmp = CreateBitmap( 8, bmpI.bmiHeader.biWidth, bmpI.bmiHeader.biHeight); break;
         case 24:    hBmp = CreateBitmap(24, bmpI.bmiHeader.biWidth, bmpI.bmiHeader.biHeight); break;
         case 32:    hBmp = CreateBitmap(32, bmpI.bmiHeader.biWidth, bmpI.bmiHeader.biHeight); break;
         default:
            //  API - LoadImage  LoadBitmap   4,8,24  32    ...
            //        .   ,        DrawBitmapEx
            _ASSERT(FALSE);
            hBmp = ::CreateCompatibleBitmap(hDC, bmpI.bmiHeader.biWidth, bmpI.bmiHeader.biHeight);
         }
         if (!hBmp)
            dwErrCode = ::GetLastError();
         ::ReleaseDC(NULL, hDC);
      }

      // 2. Set bits
      if (hBmp) {
         HDC hDC = ::CreateCompatibleDC(NULL);

         {
            HGDIOBJ  hBmpOld = ::SelectObject(hDC, hBmp);
            COLORREF clBkOld = ::SetBkColor(hDC, 0x00FFFFFF);
            COLORREF clTxOld = ::SetTextColor(hDC, 0);

            DWORD dwSize = sizeof(BITMAPINFOHEADER) + PaletteSize(bmpI.bmiHeader);
            int iLines = ::SetDIBits(hDC, hBmp, 0, bmpI.bmiHeader.biHeight, ((const BYTE*)&bmpI)+dwSize, &bmpI, DIB_RGB_COLORS);
            if (!iLines) {
               dwErrCode = ::GetLastError();
               ::DeleteObject(hBmp); hBmp = NULL;
            }

            ::SetTextColor(hDC, clTxOld);
            ::SetBkColor(hDC, clBkOld);
            ::SelectObject(hDC, hBmpOld);
         }

         if ((bmpI.bmiHeader.biBitCount == 1) && hBmp)
         //     ,    API - LoadImage  LoadBitmap
         {
            HDC hCDC = ::CreateCompatibleDC(NULL);
            BITMAP bmp;
            ::GetObject(hBmp, sizeof(bmp), &bmp);
            HBITMAP hBmp2 = ::CreateBitmap(bmp.bmWidth, bmp.bmHeight, bmp.bmPlanes, bmp.bmBitsPixel, NULL);
            if (!hBmp2) {
               dwErrCode = ::GetLastError();
               ::DeleteObject(hBmp); hBmp = NULL;
            } else {
               HGDIOBJ hBmp1Old = ::SelectObject(hDC , hBmp);
               HGDIOBJ hBmp2Old = ::SelectObject(hCDC, hBmp2);

               BOOL bRes = ::BitBlt(hCDC, 0,0,bmp.bmWidth,bmp.bmHeight, hDC, 0,0, SRCCOPY);
               if (!bRes) {
                  dwErrCode = ::GetLastError();
                  ::DeleteObject(hBmp); hBmp = NULL;
               }

               ::SelectObject(hCDC, hBmp2Old);
               ::SelectObject(hDC , hBmp1Old);

               if (bRes) {
                  HBITMAP hTemp = hBmp; hBmp = hBmp2; hBmp2 = hTemp; // hBmp <-> hBmp2
               }
            }

            ::DeleteObject(hBmp2);
            ::DeleteDC(hCDC);
         }
         ::DeleteDC(hDC);
      }
      //logBitColor(hBmp);
      ::SetLastError(dwErrCode);
      return hBmp;
   }
   HBITMAP LoadBitmap(const BYTE* pbData, DWORD nbSize) {
      ::SetLastError(NO_ERROR);
      HBITMAP hBmp = NULL;
      const BITMAPINFO *pBmpI = toBmpInfo(pbData, nbSize);
      if (pBmpI) {
         hBmp = LoadBitmap(*pBmpI);
      }
      return hBmp;
   }

   BOOL ExtractBmp(
      IN const BYTE *pbResource,
      IN DWORD       nbSize,
      IN LPCTSTR     szBmpFile
   ) {
      DWORD dwErrCode = NO_ERROR;
      BOOL bRes = ((pbResource != NULL) && (nbSize!=0));
      if (!bRes) {
         dwErrCode = ERROR_INVALID_PARAMETER;
      } else {
         HANDLE hFile = ::CreateFile(
            szBmpFile,
            GENERIC_WRITE,
            FILE_SHARE_READ,
            NULL,
            CREATE_ALWAYS,
            FILE_ATTRIBUTE_NORMAL,
            NULL
         );
         bRes = (hFile != INVALID_HANDLE_VALUE);
         if (!bRes) {
            dwErrCode = ::GetLastError();
         } else {
            const BITMAPINFO *pBmpI = toBmpInfo(pbResource, nbSize);
            bRes = !!pBmpI;
            if (!bRes) {
               dwErrCode = ::GetLastError();
            } else {
               BITMAPFILEHEADER bmpFH = {
                  WINDOWS_BITMAP_SIGNATURE,                          // WORD    bfType;  //  BMP : 0x42 = "B" 0x4d = "M"
                  sizeof(BITMAPFILEHEADER)+nbSize,                   // DWORD   bfSize;
                  0,0,                                               // WORD    bfReserved1;WORD    bfReserved2;
                  sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+ // DWORD   bfOffBits;
                     PaletteSize(pBmpI->bmiHeader)
               };
               DWORD dwNOBW = 0;
               bRes = (::WriteFile(hFile, &bmpFH    , sizeof(BITMAPFILEHEADER), &dwNOBW, NULL) && (sizeof(BITMAPFILEHEADER) == dwNOBW)) &&
                      (::WriteFile(hFile, pbResource, nbSize                  , &dwNOBW, NULL) && (nbSize                   == dwNOBW));
               if (!bRes) {
                  dwErrCode = ::GetLastError();
               }
            }
            ::CloseHandle(hFile); hFile = NULL;
         }
      }
      ::SetLastError(dwErrCode);
      return bRes;
   }

} // end namespace nsBmpInternal

SIZE SizeBitmap(LPCTSTR szBmpFile) {
   SIZE size = {0,0};
   DWORD dwErrCode = NO_ERROR;
   HANDLE hFile = ::CreateFile(
      szBmpFile,
      GENERIC_READ,
      FILE_SHARE_READ,
      NULL,
      OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL,
      NULL
   );
   BOOL bRes = (hFile != INVALID_HANDLE_VALUE);
   if (!bRes) {
      dwErrCode = ::GetLastError();
   } else {
      BITMAPFILEHEADER bmpFH;
      BITMAPINFOHEADER bmpIH;
      DWORD dwNBR = 0;
      bRes = (::ReadFile(hFile, &bmpFH, sizeof(bmpFH), &dwNBR, NULL) && (sizeof(bmpFH) == dwNBR)) &&
             (::ReadFile(hFile, &bmpIH, sizeof(bmpIH), &dwNBR, NULL) && (sizeof(bmpIH) == dwNBR));
      if (!bRes) {
         dwErrCode = ::GetLastError();
      } else {
         bRes = (bmpFH.bfType == WINDOWS_BITMAP_SIGNATURE);
         if (!bRes) {
            dwErrCode = ERROR_INVALID_DATA;
         } else {
            size.cx = bmpIH.biWidth;
            size.cy = bmpIH.biHeight;
         }
      }
      ::CloseHandle(hFile);
   }
   ::SetLastError(dwErrCode);
   return size;
}
SIZE SizeBitmap(HMODULE hMod, LPCTSTR szResourceName) { //  bitmap'      RT_BITMAP
   return SizeBitmap(hMod, szResourceName, WORD(0));
   /** /
   ::SetLastError(NO_ERROR);
   SIZE size = {0,0};
   DWORD nbSize = 0;
   const BYTE* pbResource = ::GetResource(hMod, RT_BITMAP, szResourceName, nbSize);
   if (pbResource && nbSize) {
      size = nsBmpInternal::SizeBitmap(pbResource, nbSize);
   }
   return size;
   /**/
}
SIZE SizeBitmap(HMODULE hMod, LPCTSTR szResourceName, WORD wIDLanguage) { //  bitmap'      RT_BITMAP c  
   ::SetLastError(NO_ERROR);
   SIZE size = {0,0};
   DWORD nbSize = 0;
   const BYTE* pbResource = ::GetResource(hMod, RT_BITMAP, szResourceName, wIDLanguage, nbSize);
   if (pbResource && nbSize) {
      size = nsBmpInternal::SizeBitmap(pbResource, nbSize);
   }
   return size;
}

WORD GetBitmapLang(HMODULE hMod, LPCTSTR szResourceName) { //    bitmap'      RT_BITMAP (   bitmap' )
   WORD wRes = ::GetResourceLangCount(hMod, RT_BITMAP, szResourceName);
   return wRes;
}

WORD GetBitmapCount(HMODULE hMod, BOOL bLang) { //  bitmap  DLL/EXE ( bLang==FALSE  -  RT_BITMAP). bLang -     bitmap'
   WORD wRes = ::GetResourceNameCount(hMod, RT_BITMAP, bLang);
   return wRes;
}

BOOL GetBitmapGroupNames(IN HMODULE hMod, OUT CRcName *pNames, IN WORD &wLenBuf) {
   BOOL bRes = ::GetResourceNames(hMod, RT_BITMAP, pNames, wLenBuf);
   return bRes;
}

BOOL GetBitmapGroupNamesEx(IN HMODULE hMod, OUT CRcNameEx *pNames, IN WORD &wLenBuf) {
   BOOL bRes = ::GetResourceNamesEx(hMod, RT_BITMAP, pNames, wLenBuf);
   return bRes;
}


BOOL SaveBitmap(HDC hDC, LPCTSTR szBmpFile, BOOL bReplaceFile) {
   BOOL bRes = FALSE;
   DWORD dwErrCode = NO_ERROR;

   SIZE size = {1,1};
   HBITMAP hTmpBmp = ::CreateBitmapMono(size);
   bRes = !!hTmpBmp;
   if (!bRes) {
      dwErrCode = ::GetLastError();
   } else {
      HBITMAP hBmp = (HBITMAP)::SelectObject(hDC, hTmpBmp); //     
      ::SelectObject(hDC, hBmp); // restore
      bRes = SaveBitmap(hDC, hBmp, szBmpFile, bReplaceFile);
      if (!bRes)
         dwErrCode = ::GetLastError();
      BOOL bRes2 = ::DeleteBitmap(hTmpBmp); _ASSERT(bRes2);
   }

   ::SetLastError(dwErrCode);
   return bRes;
}

BOOL SaveBitmap(HBITMAP hBmp, LPCTSTR szBmpFile, BOOL bReplaceFile) {
   BOOL bRes = FALSE;
   DWORD dwErrCode = NO_ERROR;

   HDC hDC = ::CreateCompatibleDC(NULL);
   bRes = !!hDC;
   if (!bRes) {
      dwErrCode = ::GetLastError();
   } else {
      HBITMAP hBmpOld = (HBITMAP)::SelectObject(hDC, hBmp);
      _ASSERT_EXPR(!!hBmpOld, L"  -      (    - SelectObject(hDC))");
      bRes = SaveBitmap(hDC, hBmp, szBmpFile, bReplaceFile);
      if (!bRes)
         dwErrCode = ::GetLastError();
      ::SelectObject(hDC, hBmpOld); // restore
      BOOL bRes2 = ::DeleteDC(hDC); _ASSERT(bRes2);
   }

   ::SetLastError(dwErrCode);
   return bRes;
}

BOOL SaveBitmap(HDC hDC, HBITMAP hBmp, LPCTSTR szBmpFile, BOOL bReplaceFile) {
   if (!hDC || !hBmp || !szBmpFile || !szBmpFile[0]) {
      ::SetLastError(ERROR_INVALID_PARAMETER);
      return FALSE;
   }
   BOOL bRes = FALSE;
   DWORD dwErrCode = NO_ERROR;

   HANDLE hFile = ::CreateFile(
      szBmpFile,
      GENERIC_WRITE,
      FILE_SHARE_WRITE,
      NULL,
      bReplaceFile ? CREATE_ALWAYS : CREATE_NEW,
      0,
      NULL
   );
   bRes = (hFile != INVALID_HANDLE_VALUE);
   if (!bRes) {
      dwErrCode = ::GetLastError();
   } else {
      // uncompressed file struct:
      //     BITMAPFILEHEADER
      //     BITMAPINFOHEADER
      //     palette is == 0
      //     raw bitmap data

      BITMAP bmp;
      BITMAPINFO bmi;
      ZeroMemory(&bmi, sizeof(BITMAPINFO));
      bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
      bmi.bmiHeader.biBitCount = 0;

      bRes = !!::GetDIBits(hDC, hBmp, 0, 0, NULL, &bmi, DIB_RGB_COLORS);
      if (!bRes) {
         dwErrCode = ::GetLastError();
      } else {
         bmp.bmHeight = bmi.bmiHeader.biHeight;
         bmp.bmWidth  = bmi.bmiHeader.biWidth;

         BITMAPFILEHEADER bmpFH = {
            WINDOWS_BITMAP_SIGNATURE,                                // bfType
            sizeof(BITMAPFILEHEADER)+                                // bfSize
               nsBmpInternal::BitmapSizeInBytes(bmi.bmiHeader),
            0,0,                                                     // bfReserved1 bfReserved2
            sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER)+     // bfOffBits
               nsBmpInternal::PaletteSize(bmi.bmiHeader)};

         bmi.bmiHeader.biCompression = BI_RGB; // no compression!

         DWORD dwData = nsBmpInternal::DibSectionSize(bmi.bmiHeader);
         BYTE *pData = new BYTE[dwData];
         bRes = !!pData;
         if (!bRes) {
            dwErrCode = ERROR_NOT_ENOUGH_MEMORY;
         } else {
            bRes = !!::GetDIBits(hDC, hBmp, 0, bmp.bmHeight, pData, &bmi, DIB_RGB_COLORS);
            if (!bRes) {
               dwErrCode = ::GetLastError();
            } else {
               // Writing Bitmap
               DWORD dwNBW = 0;
               bRes = (::WriteFile(hFile, &bmpFH        , sizeof(BITMAPFILEHEADER), &dwNBW, NULL) && (dwNBW == sizeof(BITMAPFILEHEADER))) &&
                      (::WriteFile(hFile, &bmi.bmiHeader, sizeof(BITMAPINFOHEADER), &dwNBW, NULL) && (dwNBW == sizeof(BITMAPINFOHEADER))) &&
                      (::WriteFile(hFile, pData         , dwData                  , &dwNBW, NULL) && (dwNBW == dwData));
               if (!bRes)
                  dwErrCode = ::GetLastError();
            }
            delete(pData); pData = NULL;
         }
      }
      ::CloseHandle(hFile);
   }

   ::SetLastError(dwErrCode);
   return bRes;
}

HBITMAP LoadBitmap(LPCTSTR szBmpFile) { //  bitmap'  
#if 0
   //    .  , ,  ,   32    
   //   32   ( -).    ....
   HBITMAP hBmp = (HBITMAP)::LoadImage(NULL, szBmpFile, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_DEFAULTCOLOR);
   return hBmp;
#else
   HBITMAP hBmp = NULL;
   DWORD dwErrCode = NO_ERROR;
   TCHAR szFile[1024] = {0};
   TCHAR *szFilePart = NULL;
   DWORD dwLen = ::SearchPath(NULL, szBmpFile, NULL, 1024, szFile, &szFilePart);
   if (dwLen == 0) {
      dwErrCode = ::GetLastError();
   } else {
      HANDLE hFile = ::CreateFile(
         szFile,
         GENERIC_READ,
         FILE_SHARE_READ,
         NULL,
         OPEN_EXISTING,
         0,
         NULL
      );
      if (hFile == INVALID_HANDLE_VALUE) {
         dwErrCode = ::GetLastError();
      } else {
         HANDLE hFileMap = ::CreateFileMapping(hFile, NULL, PAGE_READONLY, 0,0, NULL);
         if (hFileMap == 0) {
            dwErrCode = ::GetLastError();
         } else {
            LPVOID pData = ::MapViewOfFile(hFileMap, FILE_MAP_READ, 0,0,0);
            if (pData==NULL) {
               dwErrCode = ::GetLastError();
            } else {
               DWORD nbSize = ::GetFileSize(hFile, NULL);

               hBmp = LoadBitmap((const BITMAPFILEHEADER *)pData, nbSize);
               if (!hBmp) {
                  dwErrCode = ::GetLastError();
               }

               ::UnmapViewOfFile(pData);
            }
            ::CloseHandle(hFileMap);
         }
         ::CloseHandle(hFile);
      }
   }
   ::SetLastError(dwErrCode);
   return hBmp;
#endif
}

HBITMAP LoadBitmapEx(HMODULE hMod, LPCTSTR szResourceName) {
   return LoadBitmap(hMod, szResourceName, WORD(0));
   /** /
   ::SetLastError(NO_ERROR);
   DWORD nbSize = 0;
   const BYTE* pbResource = ::GetResource(hMod, RT_BITMAP, szResourceName, nbSize);
   HBITMAP hBmp = NULL;
   if (pbResource && nbSize) {
      hBmp = nsBmpInternal::LoadBitmap(pbResource, nbSize);
   }
   return hBmp;
   /**/
}

HBITMAP LoadBitmap(HMODULE hMod, LPCTSTR szResourceName, WORD wIDLanguage) { //  bitmap'      RT_BITMAP   
   ::SetLastError(NO_ERROR);
   DWORD nbSize = 0;
   const BYTE* pbResource = ::GetResource(hMod, RT_BITMAP, szResourceName, wIDLanguage, nbSize);
   HBITMAP hBmp = NULL;
   if (pbResource && nbSize) {
      hBmp = nsBmpInternal::LoadBitmap(pbResource, nbSize);
   }
   return hBmp;
}

HBITMAP LoadBitmap(const BITMAPINFO *pBmpI, DWORD nbSize) {
   if (nbSize == 0)
      nbSize = pBmpI ? nsBmpInternal::BitmapSizeInBytes(pBmpI->bmiHeader) : 0; //  
   HBITMAP hBmp = nsBmpInternal::LoadBitmap((BYTE*)pBmpI, nbSize);
   return hBmp;
}
HBITMAP LoadBitmap(const BITMAPFILEHEADER *pBmpFH, DWORD nbSize) {
   if (nbSize == 0)
      nbSize = pBmpFH ? pBmpFH->bfSize : 0; //  
   HBITMAP hBmp = NULL;
   DWORD dwErrCode = NO_ERROR;
   if (nbSize < sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)) {
      dwErrCode = ERROR_INVALID_DATA;
   } else {
      if ((pBmpFH->bfType != WINDOWS_BITMAP_SIGNATURE) ||
          (pBmpFH->bfSize != nbSize)) // ??? (pBmpFH->bfSize > nbSize))
      {
         dwErrCode = ERROR_INVALID_DATA;
      } else {
         const BITMAPINFO *pBmpI = (const BITMAPINFO *)((BYTE*)pBmpFH+sizeof(BITMAPFILEHEADER));
         hBmp = LoadBitmap(pBmpI, nbSize-sizeof(BITMAPFILEHEADER));
         if (!hBmp) {
            dwErrCode = ::GetLastError();
         }
      }
   }
   ::SetLastError(dwErrCode);
   return hBmp;
}

BOOL ExtractBmp( //  bitmap'  RT_BITMAP  bmp-
   IN HMODULE hMod,
   IN LPCTSTR szResourceName, //  bitmap'   RT_BITMAP
   IN LPCTSTR szBmpFile       //        (   !)
) {
   return ::ExtractBmp(hMod, szResourceName, WORD(0), szBmpFile);
   /** /
   ::SetLastError(NO_ERROR);
   DWORD nbSize = 0;
   const BYTE* pbResource = ::GetResource(hMod, RT_BITMAP, szResourceName, nbSize);
   BOOL bRes = (pbResource && nbSize);
   if (bRes) {
      bRes = nsBmpInternal::ExtractBmp(pbResource, nbSize, szBmpFile);
   }
   return bRes;
   /**/
}

BOOL ExtractBmp( //  bitmap'  RT_BITMAP  bmp-   
   IN HMODULE hMod,
   IN LPCTSTR szResourceName, //  bitmap'   RT_BITMAP
   IN WORD wIDLanguage,
   IN LPCTSTR szBmpFile       //        (   !)
) {
   ::SetLastError(NO_ERROR);
   DWORD nbSize = 0;
   const BYTE* pbResource = ::GetResource(hMod, RT_BITMAP, szResourceName, wIDLanguage, nbSize);
   BOOL bRes = (pbResource && nbSize);
   if (bRes) {
      bRes = nsBmpInternal::ExtractBmp(pbResource, nbSize, szBmpFile);
   }
   return bRes;
}

BOOL DrawBitmap(HDC hDCDst, RECT const &rcDst, HBITMAP hBmpSrc, RECT const &rcSrc, bool bStretch) {
   DWORD dwErrCode = NO_ERROR;

   HDC hDCSrc = ::CreateCompatibleDC(hDCDst);
   HGDIOBJ hBmpSrcOld  = ::SelectObject(hDCSrc, hBmpSrc);
   BOOL bRes = bStretch ?
      ::StretchBlt(
         hDCDst,
         rcDst.left, rcDst.top, rcDst.right-rcDst.left, rcDst.bottom-rcDst.top,
         hDCSrc,
         rcSrc.left, rcSrc.top, rcSrc.right-rcSrc.left, rcSrc.bottom-rcSrc.top,
         SRCCOPY) :
      ::BitBlt(
         hDCDst,
         rcDst.left, rcDst.top, rcDst.right-rcDst.left, rcDst.bottom-rcDst.top,
         hDCSrc,
         rcSrc.right-rcSrc.left, rcSrc.bottom-rcSrc.top,
         SRCCOPY);
   if (!bRes) {
      dwErrCode = ::GetLastError();
   }

   ::SelectObject(hDCSrc, hBmpSrcOld);
   ::DeleteDC(hDCSrc);

   ::SetLastError(dwErrCode);
   return bRes;
}
BOOL DrawBitmapAlpha(HDC hDCDst, RECT const &rcDst, HBITMAP hBmpSrc, RECT const &rcSrc) {
   BOOL bRes = FALSE;
   DWORD dwErrCode = NO_ERROR;

#ifdef _DEBUG
   BITMAP bmp;
   bRes = ::GetObject(hBmpSrc, sizeof(bmp), &bmp);
   if (bRes && (bmp.bmBitsPixel == 32)) {
#endif // _DEBUG
      SIZE sizeDst = {rcDst.right-rcDst.left, rcDst.bottom-rcDst.top};
      SIZE sizeSrc = {rcSrc.right-rcSrc.left, rcSrc.bottom-rcSrc.top};
      HDC hDCSrc = ::CreateCompatibleDC(hDCDst);//nsBmpInternal::CreateCompatibleDC(32);//
      HGDIOBJ hBmpSrcOld  = ::SelectObject(hDCSrc, hBmpSrc);

      BLENDFUNCTION blendFunction = {
         AC_SRC_OVER, // BYTE     BlendOp;
         0,           // BYTE     BlendFlags;
         255,         // BYTE     SourceConstantAlpha;
         AC_SRC_ALPHA // BYTE     AlphaFormat; (0 - ignore source alpha channel)
      };
      bRes = ::AlphaBlend(
         hDCDst,
         rcDst.left, rcDst.top, sizeDst.cx, sizeDst.cy,
         hDCSrc,
         rcSrc.left, rcSrc.top, sizeSrc.cx, sizeSrc.cy,
         blendFunction);
      if (!bRes) {
         dwErrCode = ::GetLastError();
      }

      ::SelectObject(hDCSrc, hBmpSrcOld);
      ::DeleteDC(hDCSrc);

#ifdef _DEBUG
   } else {
      _ASSERT_EXPR(FALSE, L"   -   32   ?");
   }
#endif // _DEBUG

   ::SetLastError(dwErrCode);
   return bRes;
}


BOOL DrawBitmap(HDC hDCDst, RECT const &rcDst, HBITMAP hBmpSrc, POINT const &pLTSrc) {
   BITMAP bmp;
   BOOL bRes = ::GetObject(hBmpSrc, sizeof(bmp), &bmp);
   if (bRes) {
      RECT rcSrc = {0, 0, pLTSrc.x, pLTSrc.y};
      if (bmp.bmBitsPixel == 32)
         bRes = DrawBitmapAlpha(hDCDst, rcDst, hBmpSrc, rcSrc);
      else
         bRes = DrawBitmap(hDCDst, rcDst, hBmpSrc, rcSrc, false);
   }
   return bRes;
}

BOOL DrawBitmap(HDC hDCDst, RECT const &rcDst, HBITMAP hBmpSrc, RECT const &rcSrc) {
   BITMAP bmp;
   BOOL bRes = ::GetObject(hBmpSrc, sizeof(bmp), &bmp);
   if (bRes)
      if (bmp.bmBitsPixel == 32)
         bRes = DrawBitmapAlpha(hDCDst, rcDst, hBmpSrc, rcSrc);
      else
         bRes = DrawBitmap(hDCDst, rcDst, hBmpSrc, rcSrc, true);
   return bRes;
}

///////////////////////////////
HBITMAP CreateBitmap(int iWidth, int iHeight) {
   HDC hDC = ::GetDC(NULL);
   HBITMAP hBmp = ::CreateCompatibleBitmap(hDC, iWidth, iHeight);
   ::ReleaseDC(NULL, hDC);
   return hBmp;
}

HBITMAP CreateBitmapMono(const SIZE &sizeBmp) {
   return ::CreateBitmap(sizeBmp.cx, sizeBmp.cy, 1, 1, NULL);
}

HBITMAP CreateCompatibleBitmap(HBITMAP hBitmap, int iWidth, int iHeight) {
	BITMAP bm;
	int iRes = ::GetObject(hBitmap, sizeof(bm), &bm);
   if (iRes == sizeof(BITMAP))
      return ::CreateBitmap(iWidth, iHeight, bm.bmPlanes, bm.bmBitsPixel, bm.bmBits);
   return NULL;
}

SIZE SizeBitmap(HBITMAP hBmp) {
   SIZE size = {0,0};
   if (hBmp) {
      BITMAP bmp = {0,0,0,0,0,0,NULL};
      int iRes = ::GetObject(hBmp, sizeof(BITMAP), &bmp);
      if (iRes == sizeof(BITMAP)) {
         size.cx = bmp.bmWidth;
         size.cy = bmp.bmHeight;
      }
   }
   return size;
}

SIZE SizeBitmap(HDC hDC) {
   ::SetLastError(NO_ERROR);
   SIZE size = {0,0};
   HBITMAP hBmp = (HBITMAP)::GetCurrentObject(hDC, OBJ_BITMAP);
   if (hBmp) {
      size = SizeBitmap(hBmp);
   }
   return size;
}

HBITMAP CreateBitmapMask(HBITMAP hBmp, COLORREF colorTransparent) {
   HBITMAP hBmpMask = NULL;
   DWORD dwErrCode = NO_ERROR;
   SIZE sizeBmp = ::SizeBitmap(hBmp);
   if (!sizeBmp.cx || !sizeBmp.cy) {
      dwErrCode = ::GetLastError();
   } else {
      hBmpMask = ::CreateBitmapMono(sizeBmp);
      if (!hBmpMask) {
         dwErrCode = ::GetLastError();
      } else {
         HDC hDC_Dst = ::CreateCompatibleDC(NULL);
         HDC hDC_Src = ::CreateCompatibleDC(NULL);
         HGDIOBJ hBmpSaveSrc = ::SelectObject(hDC_Src, hBmp);
         if (!hBmpSaveSrc) {
            //     hBmp     . HDC
            dwErrCode = ERROR_INVALID_OPERATION;//ERROR_RESOURCE_DISABLED;//ERROR_NO_SYSTEM_RESOURCES;//
            ::DeleteObject(hBmpMask); hBmpMask = NULL; //   ,  ...
         } else {
            HGDIOBJ hBmpSaveDst = ::SelectObject(hDC_Dst, hBmpMask);

            COLORREF oldBkColor = ::SetBkColor(hDC_Src, (colorTransparent == CLR_INVALID) ? ::GetPixel(hDC_Src, 0, 0) : colorTransparent);
            ::BitBlt(hDC_Dst, 0, 0, sizeBmp.cx, sizeBmp.cy, hDC_Src, 0, 0, NOTSRCCOPY);
            ::SetBkColor(hDC_Src, oldBkColor);

            ::SelectObject(hDC_Dst, hBmpSaveDst);
            ::SelectObject(hDC_Src, hBmpSaveSrc);
         }
         ::DeleteDC(hDC_Dst);
         ::DeleteDC(hDC_Src);
      }
   }
   ::SetLastError(dwErrCode);
   return hBmpMask;
}

HBITMAP CreateBitmapRotated(HBITMAP hBitmap, float angle, COLORREF clrBack) {
   typedef BOOL (WINAPI* PF_SetWorldTransform)(IN HDC, IN CONST XFORM *);
   static PF_SetWorldTransform func_SetWorldTransform = (PF_SetWorldTransform)::GetProcAddress(::LoadLibrary(_T("GDI32.dll")), "SetWorldTransform");

   if (func_SetWorldTransform == NULL) {
      ::SetLastError(ERROR_CALL_NOT_IMPLEMENTED);//ERROR_NOT_SUPPORTED);
      return FALSE;
   }

	// Create a memory DC compatible with the display
	HDC hDCSrc = ::CreateCompatibleDC(NULL);
   HDC hDCDst = ::CreateCompatibleDC(NULL);

	// Get logical coordinates
	BITMAP bm;
	int iRes = ::GetObject(hBitmap, sizeof(bm), &bm);

	float cosine = (float)cos(angle);
	float sine = (float)sin(angle);

	// Compute dimensions of the resulting bitmap
	// First get the coordinates of the 3 corners other than origin
	int x1 = (int)( bm.bmHeight * sine);
	int y1 = (int)( bm.bmHeight * cosine);
	int x2 = (int)( bm.bmWidth  * cosine + bm.bmHeight * sine);
	int y2 = (int)( bm.bmHeight * cosine - bm.bmWidth  * sine);
	int x3 = (int)( bm.bmWidth  * cosine);
	int y3 = (int)(-bm.bmWidth  * sine);

	int minx = min(0,min(x1, min(x2,x3)));
	int miny = min(0,min(y1, min(y2,y3)));
	int maxx = max(0,max(x1, max(x2,x3)));
	int maxy = max(0,max(y1, max(y2,y3)));

   SIZE sizeOut = {maxx - minx, maxy - miny};

	// Create a bitmap to hold the result
	HBITMAP hBmpRes = ::CreateCompatibleBitmap(hBitmap, sizeOut.cx, sizeOut.cy);
   _ASSERT(hBmpRes);

	HGDIOBJ hBmpSrcOld = ::SelectObject(hDCSrc, hBitmap);
	HGDIOBJ hBmpDstOld = ::SelectObject(hDCDst, hBmpRes);

	// Draw the background color before we change mapping mode
   {
      if (clrBack == CLR_INVALID) {
         clrBack = ::GetPixel(hDCSrc, 0,0);
      }
      HBRUSH  hBrNew = ::CreateSolidBrush(clrBack);
      HGDIOBJ hBrOld = ::SelectObject(hDCDst, hBrNew);
      BOOL bRes = ::PatBlt(hDCDst, 0, 0, sizeOut.cx, sizeOut.cy, PATCOPY);
      ::SelectObject(hDCDst, hBrOld);
      bRes = ::DeleteObject(hBrNew);
   }

	// We will use world transform to rotate the bitmap
   {
      iRes = ::SetGraphicsMode(hDCDst, GM_ADVANCED);
      XFORM xform = {
         cosine,        // FLOAT   eM11
        -sine,          // FLOAT   eM12
         sine,          // FLOAT   eM21
         cosine,        // FLOAT   eM22
         (float)-minx,  // FLOAT   eDx
         (float)-miny}; // FLOAT   eDy
	   BOOL bRes = func_SetWorldTransform(hDCDst, &xform);
   }

	// Now do the actual rotating - a pixel at a time
	BOOL bRes = ::BitBlt(hDCDst, 0,0,bm.bmWidth, bm.bmHeight, hDCSrc, 0, 0, SRCCOPY);

	// Restore DCs
	::SelectObject(hDCSrc, hBmpSrcOld);
	::SelectObject(hDCDst, hBmpDstOld);

   // free
   ::DeleteDC(hDCDst);
   ::DeleteDC(hDCSrc);

	return hBmpRes;
}

DWORD BitmapDibSectionSize(const BITMAPINFOHEADER &bih) {
   return nsBmpInternal::DibSectionSize(bih);
}
